import logging
from collections.abc import Awaitable, Callable
from functools import wraps
from typing import TYPE_CHECKING, Any, ParamSpec, TypeVar

from slixmpp import JID, Iq, Message, Presence
from slixmpp.exceptions import XMPPError
from slixmpp.xmlstream import StanzaBase

from ...contact.roster import ContactIsUser
from ...util.types import (
    AnyMUC,
    AnyRecipient,
    AnySession,
    RecipientType,
)

if TYPE_CHECKING:
    from slidge.util.types import AnyGateway


class Ignore(BaseException):
    pass


class DispatcherMixin:
    __slots__: list[str] = []

    def __init__(self, xmpp: "AnyGateway") -> None:
        self.xmpp = xmpp  # type:ignore[misc]

    async def _get_session(
        self,
        stanza: Message | Presence | Iq,
        timeout: int | None = 10,
        wait_for_ready: bool = True,
        logged: bool = False,
    ) -> AnySession:
        xmpp = self.xmpp
        if stanza.get_from().server == xmpp.boundjid.bare:
            log.debug("Ignoring echo")
            raise Ignore
        if (
            isinstance(stanza, Message)
            and stanza.get_type() == "chat"
            and stanza.get_to() == xmpp.boundjid.bare
        ):
            log.debug("Ignoring message to component")
            raise Ignore
        session = await self._get_session_from_jid(
            stanza.get_from(), timeout, wait_for_ready, logged
        )
        if isinstance(stanza, Message) and _ignore(session, stanza):
            raise Ignore
        return session

    async def _get_session_from_jid(
        self,
        jid: JID,
        timeout: int | None = 10,
        wait_for_ready: bool = True,
        logged: bool = False,
    ) -> AnySession:
        session = self.xmpp.get_session_from_jid(jid)
        if session is None:
            raise XMPPError("registration-required")
        if logged:
            session.raise_if_not_logged()
        if wait_for_ready:
            await session.wait_for_ready(timeout)
        return session

    async def get_muc_from_stanza(self, iq: Iq | Message | Presence) -> AnyMUC:
        ito = iq.get_to()
        if ito == self.xmpp.boundjid.bare:
            raise XMPPError("bad-request", text="This is only handled for MUCs")

        session = await self._get_session(iq, logged=True)
        muc = await session.bookmarks.by_jid(ito)
        return muc  # type:ignore[no-any-return]

    def _xmpp_msg_id_to_legacy(
        self,
        xmpp_id: str,
        recipient: AnyRecipient,
        origin: bool = False,
    ) -> str:
        with self.xmpp.store.session() as orm:
            sent = self.xmpp.store.id_map.get_legacy(
                orm, recipient.stored.id, xmpp_id, recipient.is_group, origin
            )
            if sent is not None:
                return sent

        return xmpp_id

    async def _get_recipient_and_thread(
        self, msg: Message
    ) -> tuple[AnyRecipient, str | None]:
        session = await self._get_session(msg)
        e: AnyRecipient = await get_recipient(session, msg)
        legacy_thread = await self._xmpp_to_legacy_thread(session, msg, e)
        return e, legacy_thread

    async def _xmpp_to_legacy_thread(
        self, session: AnySession, msg: Message, recipient: RecipientType
    ) -> str | None:
        xmpp_thread = msg["thread"]
        if not xmpp_thread:
            return None

        if session.MESSAGE_IDS_ARE_THREAD_IDS:
            return self._xmpp_msg_id_to_legacy(xmpp_thread, recipient)

        with session.xmpp.store.session() as orm:
            legacy_thread_str = session.xmpp.store.id_map.get_thread(
                orm, recipient.stored.id, xmpp_thread, recipient.is_group
            )
            if legacy_thread_str is not None:
                return legacy_thread_str
        async with session.thread_creation_lock:
            legacy_thread = await recipient.create_thread(xmpp_thread)
            with session.xmpp.store.session() as orm:
                session.xmpp.store.id_map.set_thread(
                    orm,
                    recipient.stored.id,
                    str(legacy_thread),
                    xmpp_thread,
                    recipient.is_group,
                )
                orm.commit()
            return legacy_thread


def _ignore(session: AnySession, msg: Message) -> bool:
    i = msg.get_id()
    if i.startswith("slidge-carbon-"):
        return True
    if i not in session.ignore_messages:
        return False
    session.log.debug("Ignored sent carbon: %s", i)
    session.ignore_messages.remove(i)
    return True


async def get_recipient(session: AnySession, m: Message) -> AnyRecipient:
    session.raise_if_not_logged()
    if m.get_type() == "groupchat":
        muc = await session.bookmarks.by_jid(m.get_to())
        r = m.get_from().resource
        if r not in muc.get_user_resources():
            session.create_task(muc.kick_resource(r))
            raise XMPPError("not-acceptable", "You are not connected to this chat")
        return muc  # type:ignore[no-any-return]
    else:
        return await session.contacts.by_jid(m.get_to())  # type:ignore[no-any-return]


SelfType = TypeVar("SelfType")
StanzaType = TypeVar("StanzaType", bound=StanzaBase)
HandlerType = Callable[[SelfType, StanzaType], Awaitable[None]]
P = ParamSpec("P")


def exceptions_to_xmpp_errors(cb: HandlerType[Any, Any]) -> HandlerType[Any, Any]:
    @wraps(cb)
    async def wrapped(self: SelfType, stanza: StanzaType) -> None:
        try:
            await cb(self, stanza)
        except Ignore:
            pass
        except XMPPError:
            raise
        except NotImplementedError:
            log.debug("NotImplementedError raised in %s", cb)
            raise XMPPError(
                "feature-not-implemented",
                f"{cb.__name__} is not implemented by the legacy module",
                clear=False,
            )
        except ContactIsUser:
            raise XMPPError(
                "bad-request", "Actions with your bridged self are not allowed."
            )
        except Exception as e:
            log.error(
                "Failed to handle incoming stanza: %s - %s", self, stanza, exc_info=e
            )
            raise XMPPError("internal-server-error", str(e))

    return wrapped


log = logging.getLogger(__name__)



# Commands only accessible for slidge admins
import functools
import importlib
import logging
from datetime import datetime
from typing import TYPE_CHECKING, Any, ClassVar

from slixmpp import JID
from slixmpp.exceptions import XMPPError

from ..core import config
from ..db.models import GatewayUser
from ..util.types import AnySession
from .base import (
    NODE_PREFIX,
    Command,
    CommandAccess,
    Confirmation,
    FormField,
    FormSession,
    FormValues,
    TableResult,
)
from .categories import ADMINISTRATION

NODE_PREFIX = NODE_PREFIX + "admin/"

if TYPE_CHECKING:
    from slidge.util.types import AnyGateway


class AdminCommand(Command[AnySession]):
    ACCESS = CommandAccess.ADMIN_ONLY
    CATEGORY = ADMINISTRATION


class ListUsers(AdminCommand):
    NAME = "👤 List registered users"
    HELP = "List the users registered to this gateway"
    CHAT_COMMAND = "list_users"
    NODE = NODE_PREFIX + CHAT_COMMAND

    async def run(
        self, _session: "AnySession | None", _ifrom: JID, *_: str
    ) -> TableResult:
        items = []
        with self.xmpp.store.session() as orm:
            for u in orm.query(GatewayUser).all():
                d = u.registration_date
                joined = "" if d is None else d.isoformat(timespec="seconds")
                items.append({"jid": u.jid.bare, "joined": joined})
        return TableResult(
            description="List of registered users",
            fields=[FormField("jid", type="jid-single"), FormField("joined")],
            items=items,  # type:ignore
        )


class SlidgeInfo(AdminCommand):
    NAME = "ℹ️ Server information"  # noqa:RUF001
    HELP = "List the users registered to this gateway"
    CHAT_COMMAND = "info"
    NODE = NODE_PREFIX + CHAT_COMMAND
    ACCESS = CommandAccess.ANY

    async def run(self, _session: "AnySession | None", _ifrom: JID, *_: str) -> str:
        start = self.xmpp.datetime_started
        uptime = datetime.now() - start

        if uptime.days:
            days_ago = f"{uptime.days} day{'s' if uptime.days != 1 else ''}"
        else:
            days_ago = None
        hours, seconds = divmod(uptime.seconds, 3600)

        if hours:
            hours_ago = f"{hours} hour"
            if hours != 1:
                hours_ago += "s"
        else:
            hours_ago = None

        minutes, seconds = divmod(seconds, 60)
        if minutes:
            minutes_ago = f"{minutes} minute"
            if minutes != 1:
                minutes_ago += "s"
        else:
            minutes_ago = None

        if any((days_ago, hours_ago, minutes_ago)):
            seconds_ago = None
        else:
            seconds_ago = f"{seconds} second"
            if seconds != 1:
                seconds_ago += "s"

        ago = ", ".join(
            [a for a in (days_ago, hours_ago, minutes_ago, seconds_ago) if a]
        )

        legacy_module = importlib.import_module(config.LEGACY_MODULE)
        version = getattr(legacy_module, "__version__", "No version")

        import slidge

        return (
            f"{self.xmpp.COMPONENT_NAME} (slidge core {slidge.__version__},"
            f" {config.LEGACY_MODULE} {version})\n"
            f"Up since {start:%Y-%m-%d %H:%M} ({ago} ago)"
        )


class DeleteUser(AdminCommand):
    NAME = "❌ Delete a user"
    HELP = "Unregister a user from the gateway"
    CHAT_COMMAND = "delete_user"
    NODE = NODE_PREFIX + CHAT_COMMAND

    async def run(
        self, _session: "AnySession | None", _ifrom: JID, *_: str
    ) -> FormSession[AnySession]:
        return FormSession(
            title="Remove a slidge user",
            instructions="Enter the bare JID of the user you want to delete",
            fields=[FormField("jid", type="jid-single", label="JID", required=True)],
            handler=self.delete,
        )

    async def delete(
        self,
        form_values: FormValues,
        _session: "AnySession | None",
        _ifrom: JID,
    ) -> Confirmation:
        jid: JID = form_values.get("jid")  # type:ignore
        with self.xmpp.store.session() as orm:
            user = orm.query(GatewayUser).where(GatewayUser.jid == jid).one_or_none()
        if user is None:
            raise XMPPError("item-not-found", text=f"There is no user '{jid}'")

        return Confirmation(
            prompt=f"Are you sure you want to unregister '{jid}' from slidge?",
            success=f"User {jid} has been deleted",
            handler=functools.partial(self.finish, jid=jid),
        )

    async def finish(
        self, _session: "AnySession | None", _ifrom: JID, jid: JID
    ) -> None:
        with self.xmpp.store.session() as orm:
            user = orm.query(GatewayUser).where(GatewayUser.jid == jid).one_or_none()
        if user is None:
            raise XMPPError("bad-request", f"{jid} has no account here!")
        await self.xmpp.unregister_user(user, "You have been unregistered by an admin")


class ChangeLoglevel(AdminCommand):
    NAME = "📋 Change the verbosity of the logs"
    HELP = "Set the logging level"
    CHAT_COMMAND = "loglevel"
    NODE = NODE_PREFIX + CHAT_COMMAND

    async def run(
        self, _session: "AnySession | None", _ifrom: JID, *_: str
    ) -> FormSession[AnySession]:
        return FormSession(
            title=self.NAME,
            instructions=self.HELP,
            fields=[
                FormField(
                    "level",
                    label="Log level",
                    required=True,
                    type="list-single",
                    options=[
                        {"label": "WARNING (quiet)", "value": str(logging.WARNING)},
                        {"label": "INFO (normal)", "value": str(logging.INFO)},
                        {"label": "DEBUG (verbose)", "value": str(logging.DEBUG)},
                    ],
                )
            ],
            handler=self.finish,
        )

    @staticmethod
    async def finish(
        form_values: FormValues,
        _session: "AnySession | None",
        _ifrom: JID,
    ) -> None:
        logging.getLogger().setLevel(int(form_values["level"]))  # type:ignore


class Exec(AdminCommand):
    NAME = HELP = "Exec arbitrary python code. SHOULD NEVER BE AVAILABLE IN PROD."
    CHAT_COMMAND = "!"
    NODE = "exec"
    ACCESS = CommandAccess.ADMIN_ONLY

    prev_snapshot = None

    context: ClassVar[dict[str, Any]] = {}

    def __init__(self, xmpp: "AnyGateway") -> None:
        super().__init__(xmpp)

    async def run(self, session: "AnySession | None", ifrom: JID, *args: str) -> str:
        from contextlib import redirect_stdout
        from io import StringIO

        f = StringIO()
        with redirect_stdout(f):
            exec(" ".join(args), self.context)

        out = f.getvalue()
        if out:
            return f"```\n{out}\n```"
        else:
            return "No output"



# Handle slidge commands by exchanging chat messages with the gateway components.

# Ad-hoc methods should provide a better UX, but some clients do not support them,
# so this is mostly a fallback.
import asyncio
import inspect
import logging
from collections.abc import Awaitable, Callable
from typing import (
    TYPE_CHECKING,
    Any,
    Generic,
    Literal,
    Never,
    ParamSpec,
    TypeVar,
    cast,
    overload,
)
from urllib.parse import quote as url_quote

from slixmpp import JID, CoroutineCallback, Message, StanzaPath
from slixmpp.exceptions import XMPPError
from slixmpp.types import JidStr, MessageTypes

from slidge.command.base import (
    CommandResponseRecipientType,
    CommandResponseSessionType,
    ConfirmationRecipient,
    ConfirmationSession,
    FormRecipient,
    FormSession,
)
from slidge.contact import LegacyContact
from slidge.group import LegacyMUC
from slidge.util.types import AnyMUC, AnyRecipient, AnySession

from . import Command, CommandResponseType, Confirmation, Form, TableResult
from .categories import CommandCategory

if TYPE_CHECKING:
    from ..core.gateway import BaseGateway

GatewayType = TypeVar("GatewayType", bound="BaseGateway[Any]")
T = TypeVar("T")
P = ParamSpec("P")


class ChatCommandProvider(Generic[GatewayType]):
    UNKNOWN = "Wut? I don't know that command: {}"
    xmpp: GatewayType

    def __init__(self, xmpp: GatewayType) -> None:
        self.xmpp = xmpp
        self._keywords = list[str]()
        self._commands: dict[str, Command[AnySession]] = {}
        self._input_futures = dict[str, asyncio.Future[str]]()
        self.xmpp.register_handler(
            CoroutineCallback(
                "chat_command_handler",
                StanzaPath(f"message@to={self.xmpp.boundjid.bare}"),
                self._handle_message,  # type: ignore
            )
        )

    def register(self, command: Command[AnySession]) -> None:
        """
        Register a command to be used via chat messages with the gateway

        Plugins should not call this, any class subclassing Command should be
        automatically added by slidge core.

        :param command: the new command
        """
        t = command.CHAT_COMMAND
        if t in self._commands:
            raise RuntimeError("There is already a command triggered by '%s'", t)
        self._commands[t] = command

    @overload
    async def input(self, jid: JidStr, text: str | None = None) -> str: ...

    @overload
    async def input(
        self, jid: JidStr, text: str | None = None, *, blocking: Literal[False] = ...
    ) -> asyncio.Future[str]: ...

    @overload
    async def input(
        self,
        jid: JidStr,
        text: str | None = None,
        *,
        mtype: MessageTypes = "chat",
        timeout: int = 60,
        blocking: Literal[True] = True,
        **msg_kwargs: Any,  # noqa:ANN401
    ) -> str: ...

    async def input(
        self,
        jid: JidStr,
        text: str | None = None,
        *,
        mtype: MessageTypes = "chat",
        timeout: int = 60,
        blocking: bool = True,
        **msg_kwargs: Any,
    ) -> str | asyncio.Future[str]:
        """
        Request arbitrary user input using a simple chat message, and await the result.

        You shouldn't need to call directly bust instead use :meth:`.BaseSession.input`
        to directly target a user.

        NB: When using this, the next message that the user sent to the component will
        not be transmitted to :meth:`.BaseGateway.on_gateway_message`, but rather intercepted.
        Await the coroutine to get its content.

        :param jid: The JID we want input from
        :param text: A prompt to display for the user
        :param mtype: Message type
        :param timeout:
        :param blocking: If set to False, timeout has no effect and an :class:`asyncio.Future`
            is returned instead of a str
        :return: The user's reply
        """
        jid = JID(jid)
        if text is not None:
            self.xmpp.send_message(
                mto=jid,
                mbody=text,
                mtype=mtype,
                mfrom=self.xmpp.boundjid.bare,
                **msg_kwargs,
            )
        f: asyncio.Future[str] = asyncio.get_event_loop().create_future()
        self._input_futures[jid.bare] = f
        if not blocking:
            return f
        try:
            await asyncio.wait_for(f, timeout)
        except TimeoutError:
            self.xmpp.send_message(
                mto=jid,
                mbody="You took too much time to reply",
                mtype=mtype,
                mfrom=self.xmpp.boundjid.bare,
            )
            del self._input_futures[jid.bare]
            raise XMPPError("remote-server-timeout", "You took too much time to reply")

        return f.result()

    async def _handle_message(self, msg: Message) -> None:
        if not msg["body"]:
            return

        if not msg.get_from().node:
            return  # ignore component and server messages

        f = self._input_futures.pop(msg.get_from().bare, None)
        if f is not None:
            f.set_result(msg["body"])
            return

        c = msg["body"]
        first_word, *rest = c.split(" ")
        first_word = first_word.lower()

        if first_word == "help":
            return self._handle_help(msg, *rest)

        if first_word in ("contact", "room"):
            return await self._handle_recipient(first_word, msg, *rest)

        mfrom = msg.get_from()

        command = self._commands.get(first_word)
        if command is None:
            self._not_found(msg, first_word)
            return

        try:
            session = command.raise_if_not_authorized(mfrom)
        except XMPPError as e:
            reply = msg.reply()
            reply["body"] = e.text
            reply.send()
            raise

        result: CommandResponseSessionType[Any] = await self.__wrap_handler(
            msg, command.run, session, mfrom, *rest
        )
        self.xmpp.delivery_receipt.ack(msg)
        await self._handle_result(result, msg, session)

    def __make_uri(self, body: str) -> str:
        return f"xmpp:{self.xmpp.boundjid.bare}?message;body={body}"

    async def _handle_result(
        self,
        result: CommandResponseSessionType[Any] | CommandResponseRecipientType[Any],
        msg: Message,
        session: "AnySession | None",
        recipient: AnyRecipient | None = None,
    ) -> CommandResponseSessionType[Any] | CommandResponseRecipientType[Any]:
        if isinstance(result, str) or result is None:
            reply = msg.reply()
            reply["body"] = result or "End of command."
            reply.send()
            return None

        if isinstance(result, Form):
            if recipient is None:
                result = cast(FormSession[AnySession], result)
            else:
                result = cast(FormRecipient[AnyRecipient], result)
            try:
                return await self.__handle_form(  # type:ignore[return-value]
                    result, msg, session, recipient=recipient
                )
            except XMPPError as e:
                if (
                    result.timeout_handler is None
                    or e.condition != "remote-server-timeout"
                ):
                    raise e
                return result.timeout_handler()

        if isinstance(result, Confirmation):
            yes_or_no = await self.input(msg.get_from(), result.prompt)
            if not yes_or_no.lower().startswith("y"):
                reply = msg.reply()
                reply["body"] = "Canceled"
                reply.send()
                return None
            if recipient is None:
                result = cast(ConfirmationSession[AnySession], result)
                result = await self.__wrap_handler(
                    msg,
                    result.handler,
                    session,
                    msg.get_from(),
                    *result.handler_args,
                    **result.handler_kwargs,
                )
            else:
                result = cast(ConfirmationRecipient[AnyRecipient], result)
                result = await self.__wrap_handler(
                    msg,
                    result.handler,
                    recipient,
                    *result.handler_args,
                    **result.handler_kwargs,
                )
            return await self._handle_result(result, msg, session, recipient=recipient)

        if isinstance(result, TableResult):
            if len(result.items) == 0:
                msg.reply("Empty results").send()
                return None

            body = result.description + "\n"
            for item in result.items:
                for f in result.fields:
                    if f.type == "jid-single":
                        j = JID(item[f.var])
                        value = f"xmpp:{percent_encode(j)}"
                        if result.jids_are_mucs:
                            value += "?join"
                    else:
                        value = item[f.var]  # type:ignore
                    body += f"\n{f.label or f.var}: {value}"
            msg.reply(body).send()
            return None

        raise RuntimeError

    async def __handle_form(
        self,
        result: Form,
        msg: Message,
        session: "AnySession | None",
        recipient: AnyRecipient | None = None,
    ) -> CommandResponseType:
        form_values = {}
        for t in result.title, result.instructions:
            if t:
                msg.reply(t).send()
        for f in result.fields:
            if f.type == "fixed":
                msg.reply(f"{f.label or f.var}: {f.value}").send()
            else:
                if f.type == "list-multi":
                    msg.reply(
                        "Multiple selection allowed, use new lines as a separator, ie, "
                        "one selected item per line. To select no item, reply with a space "
                        "(the punctuation)."
                    ).send()
                if f.options:
                    for o in f.options:
                        msg.reply(f"{o['label']}: {self.__make_uri(o['value'])}").send()
                if f.value:
                    msg.reply(f"Default: {f.value}").send()
                if f.type == "boolean":
                    msg.reply("yes: " + self.__make_uri("yes")).send()
                    msg.reply("no: " + self.__make_uri("no")).send()

                ans = await self.xmpp.input(
                    msg.get_from(),
                    (f.label or f.var) + "? (or 'abort')",
                    mtype="chat",
                )
                if ans.lower() == "abort":
                    return await self._handle_result("Command aborted", msg, session)
                if f.type == "boolean":
                    ans = "true" if ans.lower() == "yes" else "false"

                if f.type.endswith("multi"):
                    choices = [] if ans == " " else ans.split("\n")
                    form_values[f.var] = f.validate(choices)
                else:
                    form_values[f.var] = f.validate(ans)
        if recipient is None:
            new_result = await self.__wrap_handler(
                msg,
                result.handler,
                form_values,
                session,
                msg.get_from(),
                *result.handler_args,
                **result.handler_kwargs,
            )
            new_result = cast(CommandResponseSessionType[Any], new_result)
        else:
            new_result = await self.__wrap_handler(
                msg,
                result.handler,
                recipient,
                form_values,
                *result.handler_args,
                **result.handler_kwargs,
            )
            new_result = cast(CommandResponseRecipientType[Any], new_result)

        return await self._handle_result(new_result, msg, session, recipient=recipient)

    @staticmethod
    async def __wrap_handler(
        msg: Message,
        f: Callable[P, Awaitable[T] | T],
        *a: P.args,
        **k: P.kwargs,
    ) -> T | None:
        try:
            if inspect.iscoroutinefunction(f):
                return await f(*a, **k)  # type:ignore[no-any-return]
            elif hasattr(f, "func") and inspect.iscoroutinefunction(f.func):
                return await f(*a, **k)  # type:ignore[misc,no-any-return]
            else:
                return f(*a, **k)  # type:ignore[return-value]
        except Exception as e:
            log.debug("Error in %s", f, exc_info=e)
            reply = msg.reply()
            reply["body"] = f"Error: {e}"
            reply.send()
            return None

    def _handle_help(self, msg: Message, *rest: str) -> None:
        if len(rest) == 0:
            reply = msg.reply()
            reply["body"] = self._help(msg.get_from())
            reply.send()
        elif len(rest) == 1 and (command := self._commands.get(rest[0])):
            reply = msg.reply()
            reply["body"] = f"{command.CHAT_COMMAND}: {command.NAME}\n{command.HELP}"
            reply.send()
        else:
            self._not_found(msg, str(rest))

    def _help(self, mfrom: JID) -> str:
        session = self.xmpp.get_session_from_jid(mfrom)

        msg = "Available commands:"
        for c in sorted(
            self._commands.values(),
            key=lambda co: (
                (
                    co.CATEGORY
                    if isinstance(co.CATEGORY, str)
                    else (
                        co.CATEGORY.name
                        if isinstance(co.CATEGORY, CommandCategory)
                        else ""
                    )
                ),
                co.CHAT_COMMAND,
            ),
        ):
            try:
                c.raise_if_not_authorized(mfrom, fetch_session=False, session=session)
            except XMPPError:
                continue
            msg += f"\n{c.CHAT_COMMAND} -- {c.NAME}"
        return msg

    def _not_found(self, msg: Message, word: str) -> Never:
        e = self.UNKNOWN.format(word)
        msg.reply(e).send()
        raise XMPPError("item-not-found", e)

    async def _handle_recipient(
        self, recipient_str: Literal["contact", "room"], msg: Message, *args: str
    ) -> None:
        session = self.xmpp.get_session_from_jid(msg.get_from())

        recipient_cls = LegacyContact if recipient_str == "contact" else LegacyMUC

        if session is None:
            raise XMPPError("subscription-required")

        if len(args) == 0 or args[0] == "help":
            self.xmpp.delivery_receipt.ack(msg)
            self._help_recipient(msg, recipient_cls)
            return

        if len(args) == 1:
            self._help_recipient(msg, recipient_cls)
            raise XMPPError(
                "bad-request",
                f"Contact commands require at least two parameters: {recipient_str}_jid_username and command_name",
            )

        jid_username, command_name, *rest = args

        command = recipient_cls.commands_chat.get(command_name)
        if command is None:
            raise XMPPError("item-not-found")

        if recipient_cls is LegacyContact:
            legacy_id = await session.contacts.jid_username_to_legacy_id(jid_username)
            recipient = await session.contacts.by_legacy_id(legacy_id)
        else:
            legacy_id = await session.bookmarks.jid_username_to_legacy_id(jid_username)
            recipient = await session.bookmarks.by_legacy_id(legacy_id)

        result = await self.__wrap_handler(msg, command.run, recipient, *rest)  # type:ignore[arg-type,func-returns-value]
        self.xmpp.delivery_receipt.ack(msg)
        await self._handle_result(result, msg, session, recipient)

    def _help_recipient(
        self, msg: Message, recipient_cls: "type[LegacyContact | AnyMUC]"
    ) -> None:
        msg.reply(
            "Available commands:\n"
            + "\n".join(
                f"{co.CHAT_COMMAND} ({co.NAME}): {co.HELP}"
                for co in recipient_cls.commands_chat.values()
            )
        ).send()


def percent_encode(jid: JID) -> str:
    return f"{url_quote(jid.user)}@{jid.server}"


log = logging.getLogger(__name__)



from slixmpp.exceptions import XMPPError

from slidge import (
    BaseGateway,
    BaseSession,
    LegacyBookmarks,
    LegacyContact,
    LegacyParticipant,
    LegacyRoster,
)
from slidge.group import LegacyMUC
from slidge.util.test import SlidgeTest


class Session(BaseSession):
    async def login(self):
        pass


class Gateway(BaseGateway):
    COMPONENT_NAME = "Megacorp Chat XMPP Gateway"
    GROUPS = True


class Roster(LegacyRoster[LegacyContact]):
    async def legacy_id_to_jid_username(self, legacy_id: str) -> str:
        if legacy_id.startswith("room-"):
            raise XMPPError("item-not-found", "This is a MUC JID")
        return legacy_id


class Bookmarks(LegacyBookmarks[LegacyMUC[LegacyParticipant]]):
    async def fill(self) -> None:
        pass


class TestOccupantId(SlidgeTest):
    plugin = globals()

    def setUp(self) -> None:
        super().setUp()
        self._add_slidge_user()

    def test_update_contact_in_previously_anonymous_participant(self) -> None:
        muc = self.get_joined_muc("room-occupant")
        self.run_coro(muc.get_participant(occupant_id="occupant-id"))
        self.run_coro(
            muc.get_participant_by_legacy_id("user-id", occupant_id="occupant-id")
        )

        by_occupant_id: LegacyParticipant = self.run_coro(
            muc.get_participant(occupant_id="occupant-id")
        )
        by_legacy_id: LegacyParticipant = self.run_coro(
            muc.get_participant_by_legacy_id("user-id")
        )

        assert by_occupant_id.occupant_id == "occupant-id"
        assert by_legacy_id.occupant_id == "occupant-id"

        assert by_occupant_id.contact.legacy_id == "user-id"
        assert by_legacy_id.contact.legacy_id == "user-id"



from typing import TYPE_CHECKING

from slixmpp import Message
from slixmpp.xmlstream import StanzaBase

from ..util import DispatcherMixin, exceptions_to_xmpp_errors

if TYPE_CHECKING:
    from slidge.util.types import AnyGateway


class ChatStateMixin(DispatcherMixin):
    __slots__: list[str] = []

    def __init__(self, xmpp: "AnyGateway") -> None:
        super().__init__(xmpp)
        xmpp.add_event_handler("chatstate", self.on_chatstate)

    @exceptions_to_xmpp_errors
    async def on_chatstate(self, msg: StanzaBase) -> None:
        assert isinstance(msg, Message)
        if msg["body"]:
            # if there is a body, it's handled in on_legacy_message()
            return
        recipient, thread = await self._get_recipient_and_thread(msg)
        await recipient.on_chat_state(msg["chat_state"], thread)



import abc

from slixmpp import JID
from slixmpp.plugins.xep_0004.stanza.form import Form

from ...util.types import ChatState, Sticker, XMPPMessage


class ReactionMixin:
    REACTIONS_SINGLE_EMOJI = False

    async def restricted_emoji_extended_feature(self) -> Form | None:
        available = await self.available_emojis()
        if not self.REACTIONS_SINGLE_EMOJI and available is None:
            return None

        form = Form()
        form["type"] = "result"
        form.add_field("FORM_TYPE", "hidden", value="urn:xmpp:reactions:0:restrictions")
        if self.REACTIONS_SINGLE_EMOJI:
            form.add_field("max_reactions_per_user", value="1", type="text-single")
        if available:
            form.add_field("allowlist", value=list(available), type="text-multi")
        return form

    @abc.abstractmethod
    async def available_emojis(
        self, legacy_msg_id: str | None = None
    ) -> set[str] | None:
        """
        Override this to restrict the subset of reactions this recipient
        can handle.

        :return: A set of emojis or None if any emoji is allowed
        """
        return None


class ThreadMixin:
    @abc.abstractmethod
    async def create_thread(self, xmpp_id: str) -> str:
        return xmpp_id


class MessageMixin:
    @abc.abstractmethod
    async def on_message(self, message: XMPPMessage) -> str | None:
        """
        Triggered when the user sends a message to this :term:`Recipient`.
        """
        raise NotImplementedError

    @abc.abstractmethod
    async def on_sticker(self, sticker: Sticker) -> str | None:
        """
        Triggered when the user sends a sticker to this :term:`Recipient`.

        :param sticker: The sticker sent by the user.

        :return: An ID of some sort that can be used later to ack and mark the
            message as read by the user
        """
        raise NotImplementedError

    async def on_chat_state(self, chat_state: ChatState, thread: str | None) -> None:
        """
        Triggered when the user sends a chat state (:xep:`0085`) to this
        :term:`Recipient`.

        :param chat_state: The sticker sent by the user.
        """
        raise NotImplementedError

    async def on_displayed(self, legacy_msg_id: str, thread: str | None) -> None:
        """
        Triggered when the user sends a read marker (:xep:`0333`) to this
        :term:`Recipient`.

        This is only possible if a valid ``legacy_msg_id`` was passed when
        transmitting a message from a legacy chat to the user, eg in
        :meth:`slidge.contact.LegacyContact.send_text`
        or
        :meth:`slidge.group.LegacyParticipant.send_text`.

        :param legacy_msg_id: Identifier of the message/
        :param thread:
        """
        raise NotImplementedError

    async def on_react(
        self, legacy_msg_id: str, emojis: list[str], thread: str | None
    ) -> None:
        """
        Triggered when the user sends an emoji reaction (:xep:`0444`) to
        this :term:`Recipient`.

        :param legacy_msg_id: ID of the message the user reacts to
        :param emojis: Unicode characters representing reactions to the message ``legacy_msg_id``.
            An empty list means "no reaction", ie, remove all reactions if any were present before
        :param thread:
        """
        raise NotImplementedError

    async def on_retract(self, legacy_msg_id: str, thread: str | None) -> None:
        """
        Triggered when the user retracts (:xep:`0424`) a message.

        :param legacy_msg_id: Legacy ID of the retracted message
        :param thread:
        """
        raise NotImplementedError


class RecipientMixin(ReactionMixin, ThreadMixin, MessageMixin):
    jid: JID

    def __eq__(self, other: object) -> bool:
        return isinstance(other, self.__class__) and self.jid == other.jid



from slixmpp import JID, Iq

from slidge import BaseGateway, BaseSession, GatewayUser
from slidge.contact import LegacyContact
from slidge.util.test import SlidgeTest


class Session(BaseSession):
    async def login(self):
        pass


class Gateway(BaseGateway):
    COMPONENT_NAME = "A test"


class Contact(LegacyContact):
    REACTIONS_SINGLE_EMOJI = True

    async def on_react(
        self,
        legacy_msg_id: str,
        emojis: list[str],
        thread: str | None = None,
    ):
        pass


class TestGatewayWideRestrictions(SlidgeTest):
    plugin = globals()
    xmpp: Gateway

    def setUp(self):
        super().setUp()
        with self.xmpp.store.session() as orm:
            user = GatewayUser(
                jid=JID("romeo@montague.lit/gajim").bare,
                legacy_module_data={"username": "romeo", "city": ""},
                preferences={"sync_avatar": True, "sync_presence": True},
            )
            orm.add(user)
            orm.commit()
        self.run_coro(
            self.xmpp._BaseGateway__dispatcher._on_user_register(
                Iq(sfrom="romeo@montague.lit/gajim")
            )
        )
        welcome = self.next_sent()
        assert welcome["body"]
        stanza = self.next_sent()
        assert "logging in" in stanza["status"].lower(), stanza
        stanza = self.next_sent()
        assert "syncing contacts" in stanza["status"].lower(), stanza
        probe = self.next_sent()
        assert probe.get_type() == "probe"
        stanza = self.next_sent()
        assert "logged in" in stanza["status"].lower(), stanza

        self.send(  # language=XML
            """
            <iq type="get"
                to="romeo@montague.lit"
                id="1"
                from="aim.shakespeare.lit">
              <pubsub xmlns="http://jabber.org/protocol/pubsub">
                <items node="urn:xmpp:avatar:metadata" />
              </pubsub>
            </iq>
            """
        )

    @property
    def romeo_session(self) -> Session:
        return BaseSession.get_self_or_unique_subclass().from_jid(
            JID("romeo@montague.lit")
        )

    def test_gateway_wide_restrictions(self):
        self.recv(  # language=XML
            f"""
            <iq from='romeo@montague.lit'
                to='{self.xmpp.boundjid.bare}'
                type='get'>
              <query xmlns='http://jabber.org/protocol/disco#info' />
            </iq>
            """
        )
        self.send(  # language=XML
            """
            <iq xmlns="jabber:component:accept"
                from="aim.shakespeare.lit"
                to="romeo@montague.lit"
                type="result"
                id="2">
              <query xmlns="http://jabber.org/protocol/disco#info">
                <identity category="account"
                          type="registered"
                          name="A test" />
                <identity category="pubsub"
                          type="pep"
                          name="A test" />
                <identity category="gateway"
                          type=""
                          name="A test" />
                <feature var="jabber:iq:search" />
                <feature var="jabber:iq:register" />
                <feature var="urn:ietf:params:xml:ns:vcard-4.0" />
                <feature var="urn:xmpp:mds:displayed:0+notify" />
                <feature var="urn:xmpp:mds:displayed:0" />
                <feature var="http://jabber.org/protocol/pubsub#event" />
                <feature var="http://jabber.org/protocol/pubsub#retrieve-items" />
                <feature var="http://jabber.org/protocol/pubsub#persistent-items" />
                <feature var="urn:xmpp:avatar:metadata+notify" />
                <feature var="urn:xmpp:chat-markers:0" />
                <feature var="jabber:iq:gateway" />
                <feature var="urn:xmpp:ping" />
                <feature var="http://jabber.org/protocol/commands" />
                <feature var="urn:xmpp:pubsub-filter:0" />
                <x xmlns="jabber:x:data"
                   type="result">
                  <field var="FORM_TYPE"
                         type="hidden">
                    <value>urn:xmpp:reactions:0:restrictions</value>
                  </field>
                  <field var="max_reactions_per_user"
                         type="text-single">
                    <value>1</value>
                  </field>
                  <field var="scope">
                    <value>domain</value>
                  </field>
                </x>
              </query>
            </iq>
            """
        )

    def test_error_multi_emoji(self):
        # one of the other tests has a persisting side effect which is triggered only
        # when running the full suite. :(
        self.xmpp.plugin["xep_0356"].granted_privileges.clear()
        self.recv(  # language=XML
            f"""
            <message from='romeo@montague.lit'
                     to='juliet@{self.xmpp.boundjid.bare}'>
              <reactions id='some-id'
                         xmlns='urn:xmpp:reactions:0'>
                <reaction>👋</reaction>
                <reaction>🐢</reaction>
              </reactions>
            </message>
            """
        )
        self.send(  # language=XML
            """
            <message xmlns="jabber:component:accept"
                     type="chat"
                     from="aim.shakespeare.lit"
                     to="romeo@montague.lit">
              <body>Maximum 1 emoji/message</body>
              <active xmlns="http://jabber.org/protocol/chatstates" />
              <markable xmlns="urn:xmpp:chat-markers:0" />
              <store xmlns="urn:xmpp:hints" />
            </message>
            """
        )
        self.send(  # language=XML
            """
            <message from="juliet@aim.shakespeare.lit"
                     to="romeo@montague.lit"
                     type="error">
              <reactions xmlns="urn:xmpp:reactions:0"
                         id="some-id">
                <reaction>👋</reaction>
                <reaction>🐢</reaction>
              </reactions>
              <error type="modify">
                <feature-not-implemented xmlns="urn:ietf:params:xml:ns:xmpp-stanzas" />
                <text xmlns="urn:ietf:params:xml:ns:xmpp-stanzas">Maximum 1 emoji/message</text>
              </error>
            </message>
            """,
            use_values=False,
        )



import logging
import mimetypes
import re
from collections.abc import Callable, Collection, Coroutine
from functools import wraps
from pathlib import Path
from time import time
from typing import (
    Any,
    ClassVar,
    Concatenate,
    NamedTuple,
    ParamSpec,
    Protocol,
    TypeVar,
)

try:
    import emoji
except ImportError:
    EMOJI_LIB_AVAILABLE = False
else:
    EMOJI_LIB_AVAILABLE = True

from slixmpp.types import ExtPresenceShows, ResourceDict

from .types import Mention

try:
    import magic
except ImportError as e:
    magic = None  # type:ignore
    logging.warning(
        (
            "Libmagic is not available: %s. "
            "It's OK if you don't use fix-filename-suffix-mime-type."
        ),
        e,
    )


def fix_suffix(
    path: Path, mime_type: str | None, file_name: str | None
) -> tuple[str, str]:
    guessed = magic.from_file(path, mime=True)
    if guessed == mime_type:
        log.debug("Magic and given MIME match")
    else:
        log.debug("Magic (%s) and given MIME (%s) differ", guessed, mime_type)
        mime_type = guessed

    valid_suffix_list = mimetypes.guess_all_extensions(mime_type, strict=False)

    name = Path(file_name) if file_name else Path(path.name)

    suffix = name.suffix

    if suffix in valid_suffix_list:
        log.debug("Suffix %s is in %s", suffix, valid_suffix_list)
        return str(name), guessed

    valid_suffix = mimetypes.guess_extension(mime_type.split(";")[0], strict=False)
    if valid_suffix is None:
        log.debug("No valid suffix found")
        return str(name), guessed

    log.debug("Changing suffix of %s to %s", file_name or path.name, valid_suffix)
    return str(name.with_suffix(valid_suffix)), guessed


class SubclassableOnce:
    # To allow importing everything, including plugins, during tests
    TEST_MODE: bool = False
    __subclasses: ClassVar[
        dict[type["SubclassableOnce"], type["SubclassableOnce"] | None]
    ] = {}

    def __init_subclass__(cls, **kwargs: object) -> None:
        if SubclassableOnce not in cls.__bases__:
            base = SubclassableOnce.__find_direct_child(cls)
            existing = SubclassableOnce.__subclasses.get(base)
            if existing is not None and not SubclassableOnce.TEST_MODE:
                raise RuntimeError("This class must be subclassed once at most!")
            cls.__subclasses[base] = cls
        super().__init_subclass__(**kwargs)

    @staticmethod
    def __find_direct_child(cls: type["SubclassableOnce"]) -> type["SubclassableOnce"]:
        for base in cls.__bases__:
            if issubclass(base, SubclassableOnce):
                return base
        else:
            raise RuntimeError("wut")

    @classmethod
    def get_self_or_unique_subclass(cls) -> "type[SubclassableOnce]":
        try:
            return cls.get_unique_subclass()
        except AttributeError:
            return cls

    @classmethod
    def get_unique_subclass(cls) -> "type[SubclassableOnce]":
        existing = SubclassableOnce.__subclasses.get(cls)
        if existing is None:
            raise AttributeError("Could not find any subclass", cls)
        return existing

    @classmethod
    def reset_subclass(cls) -> None:
        log.debug("Resetting subclass of %s", cls)
        cls.__subclasses[cls] = None


def is_valid_phone_number(phone: str | None) -> bool:
    if phone is None:
        return False
    match = re.match(r"\+\d.*", phone)
    if match is None:
        return False
    return match[0] == phone


def strip_illegal_chars(s: str, repl: str = "") -> str:
    return ILLEGAL_XML_CHARS_RE.sub(repl, s)


# from https://stackoverflow.com/a/64570125/5902284 and Link Mauve
ILLEGAL = [
    (0x00, 0x08),
    (0x0B, 0x0C),
    (0x0E, 0x1F),
    (0x7F, 0x84),
    (0x86, 0x9F),
    (0xFDD0, 0xFDDF),
    (0xFFFE, 0xFFFF),
    (0x1FFFE, 0x1FFFF),
    (0x2FFFE, 0x2FFFF),
    (0x3FFFE, 0x3FFFF),
    (0x4FFFE, 0x4FFFF),
    (0x5FFFE, 0x5FFFF),
    (0x6FFFE, 0x6FFFF),
    (0x7FFFE, 0x7FFFF),
    (0x8FFFE, 0x8FFFF),
    (0x9FFFE, 0x9FFFF),
    (0xAFFFE, 0xAFFFF),
    (0xBFFFE, 0xBFFFF),
    (0xCFFFE, 0xCFFFF),
    (0xDFFFE, 0xDFFFF),
    (0xEFFFE, 0xEFFFF),
    (0xFFFFE, 0xFFFFF),
    (0x10FFFE, 0x10FFFF),
]

ILLEGAL_RANGES = [rf"{chr(low)}-{chr(high)}" for (low, high) in ILLEGAL]
XML_ILLEGAL_CHARACTER_REGEX = "[" + "".join(ILLEGAL_RANGES) + "]"
ILLEGAL_XML_CHARS_RE = re.compile(XML_ILLEGAL_CHARACTER_REGEX)


# from https://stackoverflow.com/a/35804945/5902284
def addLoggingLevel(
    levelName: str = "TRACE",
    levelNum: int = logging.DEBUG - 5,
    methodName: str | None = None,
) -> None:
    """
    Comprehensively adds a new logging level to the `logging` module and the
    currently configured logging class.

    `levelName` becomes an attribute of the `logging` module with the value
    `levelNum`. `methodName` becomes a convenience method for both `logging`
    itself and the class returned by `logging.getLoggerClass()` (usually just
    `logging.Logger`). If `methodName` is not specified, `levelName.lower()` is
    used.

    To avoid accidental clobberings of existing attributes, this method will
    raise an `AttributeError` if the level name is already an attribute of the
    `logging` module or if the method name is already present

    Example
    -------
    >>> addLoggingLevel('TRACE', logging.DEBUG - 5)
    >>> logging.getLogger(__name__).setLevel("TRACE")
    >>> logging.getLogger(__name__).trace('that worked')
    >>> logging.trace('so did this')
    >>> logging.TRACE
    5

    """
    if not methodName:
        methodName = levelName.lower()

    if hasattr(logging, levelName):
        log.debug(f"{levelName} already defined in logging module")
        return
    if hasattr(logging, methodName):
        log.debug(f"{methodName} already defined in logging module")
        return
    if hasattr(logging.getLoggerClass(), methodName):
        log.debug(f"{methodName} already defined in logger class")
        return

    # This method was inspired by the answers to Stack Overflow post
    # http://stackoverflow.com/q/2183233/2988730, especially
    # http://stackoverflow.com/a/13638084/2988730
    def logForLevel(self, message, *args, **kwargs) -> None:  # type:ignore[no-untyped-def]  # noqa
        if self.isEnabledFor(levelNum):
            self._log(levelNum, message, args, **kwargs)

    def logToRoot(message, *args, **kwargs) -> None:  # type:ignore[no-untyped-def]  # noqa
        logging.log(levelNum, message, *args, **kwargs)

    logging.addLevelName(levelNum, levelName)
    setattr(logging, levelName, levelNum)
    setattr(logging.getLoggerClass(), methodName, logForLevel)
    setattr(logging, methodName, logToRoot)


class SlidgeLogger(logging.Logger):
    def trace(self) -> None:
        pass


log = logging.getLogger(__name__)


def merge_resources(resources: dict[str, ResourceDict]) -> ResourceDict | None:
    if len(resources) == 0:
        return None

    if len(resources) == 1:
        return next(iter(resources.values()))

    by_priority = sorted(resources.values(), key=lambda r: r["priority"], reverse=True)

    if any(r["show"] == "" for r in resources.values()):
        # if a client is "available", we're "available"
        show: ExtPresenceShows = ""
    else:
        for r in by_priority:
            if r["show"]:
                show = r["show"]
                break
        else:
            raise RuntimeError()

    # if there are different statuses, we use the highest priority one,
    # but we ignore resources without status, even with high priority
    status = ""
    for r in by_priority:
        if r["status"]:
            status = r["status"]
            break

    return {
        "show": show,
        "status": status,
        "priority": 0,
    }


_EMOJI_VARIATION_SELECTOR = "\ufe0f"


def remove_emoji_variation_selector_16(emoji: str) -> str:
    # this is required for compatibility with dino, and maybe other future clients?
    return emoji.removesuffix(_EMOJI_VARIATION_SELECTOR)


NamedTupleT = TypeVar("NamedTupleT", bound=NamedTuple)


def dict_to_named_tuple(data: dict[str, Any], cls: type[NamedTupleT]) -> NamedTupleT:
    return cls(*(data.get(f) for f in cls._fields))  # type:ignore[arg-type]


def replace_mentions(
    text: str,
    mentions: Collection[Mention] | None,
    mapping: Callable[[Mention], str],
) -> str:
    if not mentions:
        return text

    cursor = 0
    pieces = []
    for mention in mentions:
        try:
            new_text = mapping(mention)
        except Exception as exc:
            log.debug("Attempting slidge <= 0.3.3 compatibility: %s", exc)
            new_text = mapping(mention.contact)  # type:ignore
        pieces.extend([text[cursor : mention.start], new_text])
        cursor = mention.end
    pieces.append(text[cursor:])
    return "".join(pieces)


class HasLogger(Protocol):
    log: logging.Logger


P = ParamSpec("P")
T = TypeVar("T")
Self = TypeVar("Self", bound=HasLogger)
TimeItWrapped = Callable[Concatenate[Self, P], Coroutine[Any, Any, T]]


def timeit(func: TimeItWrapped[Self, P, T]) -> TimeItWrapped[Self, P, T]:
    @wraps(func)
    async def wrapped(self: Self, /, *args: P.args, **kwargs: P.kwargs) -> T:
        start = time()
        r = await func(self, *args, **kwargs)
        self.log.debug("%s took %s ms", func.__name__, round((time() - start) * 1000))
        return r

    return wrapped


def strip_leading_emoji(text: str) -> str:
    if not EMOJI_LIB_AVAILABLE:
        return text
    words = text.split(" ")
    # is_emoji returns False for 🛷️ for obscure reasons,
    # purely_emoji seems better
    if len(words) > 1 and emoji.purely_emoji(words[0]):
        return " ".join(words[1:])
    return text


async def noop_coro() -> None:
    pass


def add_quote_prefix(text: str) -> str:
    """
    Return multi-line text with leading quote marks (i.e. the ">" character).
    """
    return "\n".join(("> " + x).strip() for x in text.split("\n")).strip()



from slixmpp import CoroutineCallback, Iq, StanzaPath
from slixmpp.exceptions import XMPPError
from slixmpp.plugins.xep_0060.stanza import Affiliation, Item, Subscription
from slixmpp.plugins.xep_0084.stanza import Data as AvatarData
from slixmpp.plugins.xep_0084.stanza import MetaData as AvatarMetadata

from slidge.db.models import Space

from ...util.types import AnyGateway, AnySession
from .util import DispatcherMixin, exceptions_to_xmpp_errors


class PubSubMixin(DispatcherMixin):
    __slots__: list[str] = []

    def __init__(self, xmpp: AnyGateway) -> None:
        super().__init__(xmpp)

        for path, func in [
            ("get/pubsub/items", self.__get_items),
            ("set/pubsub/subscribe", self.__set_subscribe),
            ("set/pubsub/unsubscribe", self.__set_unsubscribe),
            ("get/pubsub/affiliations", self.__get_affiliations),
            ("get/pubsub_owner/affiliations", self.__owner_get_affiliations),
            ("get/pubsub/subscriptions", self.__get_subscriptions),
            ("get/pubsub_owner/subscriptions", self.__owner_get_subscriptions),
        ]:
            self.xmpp.register_handler(
                CoroutineCallback(
                    func.__name__,
                    StanzaPath(f"iq@to={self.xmpp.boundjid.bare}@type={path}"),
                    func,  # type:ignore[arg-type]  # ty:ignore[invalid-argument-type]
                )
            )

    async def __get_items(self, iq: Iq) -> None:
        node = iq["pubsub"]["items"]["node"]

        if node in (AvatarData.namespace, AvatarMetadata.namespace, "urn:xmpp:vcard4"):
            # handled in slidge/core/pubsub.py
            # TODO: have a single entrypoint for all get/pubsub/items
            return

        session = await self._get_session(iq, logged=True)
        legacy_id = await self.__get_legacy_id(session, node)
        item_ids = [item["id"] for item in iq["pubsub"]["items"]["items"]]

        # TODO: support for avatar and banner items here
        if not all(
            item_id.endswith(f"@{self.xmpp.boundjid.bare}") for item_id in item_ids
        ):
            raise XMPPError(
                "item-not-found",
                "Slidge only supports items that are MUCs.",
            )

        if item_ids:
            room_legacy_ids = [
                str(
                    await session.bookmarks.jid_local_part_to_legacy_id(x.split("@")[0])
                )
                for x in item_ids
            ]
        else:
            room_legacy_ids = []

        with self.xmpp.store.session() as orm:
            rooms = self.xmpp.store.spaces.get_rooms(
                orm, session.user_pk, str(legacy_id), room_legacy_ids
            )

        if item_ids and len(rooms) != len(item_ids):
            raise XMPPError(
                "item-not-found",
                f"Could not find items: {set(item_ids) - set(str(r.jid) for r in rooms)}",
            )

        reply = iq.reply()
        reply["pubsub"]["items"]["node"] = node
        for room in rooms:
            item = Item()
            item["id"] = room.jid
            item["conference"]["name"] = room.name
            reply["pubsub"]["items"].append(item)

        reply.send()

    async def __get_legacy_id(self, session: AnySession, node: str) -> object:
        try:
            legacy_id = await session.bookmarks.space_node_to_legacy_id(node)
        except Exception as e:
            raise XMPPError("item-not-found", str(e))
        with self.xmpp.store.session() as orm:
            if not self.xmpp.store.spaces.exists(orm, session.user_pk, str(legacy_id)):
                raise XMPPError(
                    "item-not-found", f"This is not a known space: '{legacy_id}'"
                )
        return legacy_id

    async def __set_subscribe(self, iq: Iq) -> None:
        node = iq["pubsub"]["subscribe"]["node"]

        session = await self._get_session(iq, logged=True)
        await self.__get_legacy_id(session, node)

        reply = iq.reply(clear=True)
        sub = reply["pubsub"]["subscription"]
        sub["node"] = node
        sub["jid"] = session.user_jid
        sub["subscription"] = "subscribed"
        reply.send()

    @exceptions_to_xmpp_errors
    async def __set_unsubscribe(self, iq: Iq) -> None:
        session = await self._get_session(iq, logged=True)

        jid = iq["pubsub"]["unsubscribe"]["jid"]
        if jid != session.user_jid.bare:
            raise XMPPError("bad-request", f"Cannot unsubscribe JID {jid}")

        node = iq["pubsub"]["unsubscribe"]["node"]
        legacy_id = await session.bookmarks.space_node_to_legacy_id(node)

        await session.on_leave_space(legacy_id)
        reply = iq.reply(clear=True)

        sub = reply["pubsub"]["subscription"]
        sub["node"] = node
        sub["jid"] = session.user_jid
        sub["subscription"] = "none"
        reply.send()

    async def __get_affiliations(self, iq: Iq) -> None:
        session = await self._get_session(iq, logged=True)

        reply = iq.reply()
        node = iq["pubsub"]["affiliations"]["node"]
        if node:
            legacy_id = await self.__get_legacy_id(session, node)
            with self.xmpp.store.session() as orm:
                if not self.xmpp.store.spaces.exists(
                    orm, session.user_pk, str(legacy_id)
                ):
                    raise XMPPError("item-not-found", f"Space '{legacy_id}' not found")
            affiliation = reply["pubsub"]["affiliations"]["affiliation"]
            affiliation["node"] = node
            affiliation["affiliation"] = "subscriber"
        else:
            with self.xmpp.store.session() as orm:
                spaces = list(self.xmpp.store.spaces.get_all(orm, session.user_pk))
            for space in spaces:
                reply["pubsub"]["affiliations"].append(
                    await self.__make_affiliation(session, space)
                )

        reply.send()

    @staticmethod
    async def __make_affiliation(session: AnySession, space: Space) -> Affiliation:
        affiliation = Affiliation()
        node = await session.bookmarks.space_legacy_id_to_node(space.legacy_id)
        affiliation["node"] = node
        affiliation["affiliation"] = "subscriber"
        return affiliation

    async def __owner_get_affiliations(self, iq: Iq) -> None:
        raise XMPPError(
            "forbidden", "Slidge does not implement managing space affiliations."
        )

    async def __owner_get_subscriptions(self, iq: Iq) -> None:
        node = iq["pubsub_owner"]["subscriptions"]["node"]
        if not node:
            raise XMPPError("bad-request", "No node was specified")

        session = await self._get_session(iq, logged=True)
        await self.__get_legacy_id(session, node)

        reply = iq.reply(clear=False)
        sub = reply["pubsub_owner"]["subscriptions"]["subscription"]
        sub["jid"] = session.user_jid
        sub["subscription"] = "subscribed"
        reply.send()

    async def __get_subscriptions(self, iq: Iq) -> None:
        _node = iq["pubsub"]["subscriptions"]["node"]

        session = await self._get_session(iq, logged=True)

        reply = iq.reply(clear=False)
        subscriptions = reply["pubsub"]["subscriptions"]
        if node := iq["pubsub"]["subscriptions"]["node"]:
            await self.__get_legacy_id(session, node)
            subscription = subscriptions["subscription"]
            subscription["node"] = node
            subscription["jid"] = session.user_jid
            subscription["subscription"] = "subscribed"
        else:
            with self.xmpp.store.session() as orm:
                spaces = list(self.xmpp.store.spaces.get_all(orm, session.user_pk))
            for space in spaces:
                subscriptions.append(await self.__make_subscription(session, space))

        reply.send()

    @staticmethod
    async def __make_subscription(session: AnySession, space: Space) -> Subscription:
        subscription = Subscription()
        node = await session.bookmarks.space_legacy_id_to_node(space.legacy_id)
        subscription["node"] = node
        subscription["jid"] = session.user_jid
        subscription["subscription"] = "subscribed"
        return subscription
